@@ -25,7 +25,7 @@ module Data.Vector.Fusion.Stream.Monadic (
   empty, singleton, cons, snoc, replicate, replicateM, generate, generateM, (++),
 
   -- * Accessing elements
-  head, last, (!!),
+  head, last, (!!), (!?),
 
   -- * Substreams
   slice, init, tail, take, drop,
@@ -56,6 +56,7 @@ module Data.Vector.Fusion.Stream.Monadic (
   -- * Unfolding
   unfoldr, unfoldrM,
   unfoldrN, unfoldrNM,
+  iterateN, iterateNM,
 
   -- * Scans
   prescanl, prescanlM, prescanl', prescanlM',
@@ -254,6 +255,7 @@ last (Stream step s _) = last_loop0 SPEC s
             Skip    s' -> last_loop1 SPEC x s'
             Done       -> return x
 
+infixl 9 !!
 -- | Element at the given position
 (!!) :: Monad m => Stream m a -> Int -> m a
 {-# INLINE (!!) #-}
@@ -270,6 +272,22 @@ Stream step s _ !! i | i < 0     = BOUNDS_ERROR(error) "!!" "negative index"
             Skip    s'             -> index_loop SPEC s' i
             Done                   -> BOUNDS_ERROR(emptyStream) "!!"
 
+infixl 9 !?
+-- | Element at the given position or 'Nothing' if out of bounds
+(!?) :: Monad m => Stream m a -> Int -> m (Maybe a)
+{-# INLINE (!?) #-}
+Stream step s _ !? i = index_loop SPEC s i
+  where
+    index_loop !sPEC s i
+      = i `seq`
+        do
+          r <- step s
+          case r of
+            Yield x s' | i == 0    -> return (Just x)
+                       | otherwise -> index_loop SPEC s' (i-1)
+            Skip    s'             -> index_loop SPEC s' i
+            Done                   -> return Nothing
+
 -- Substreams
 -- ----------
 
@@ -356,7 +374,6 @@ drop n (Stream step s sz) = Stream step' (s, Just n) (sz - Exact n)
                              Done       -> Done
                            ) (step s)
                      
-
 -- Mapping
 -- -------
 
@@ -943,7 +960,7 @@ flatten mk istep sz (Stream ostep t _) = Stream step (Left t) sz
                       case r of
                         Yield a t' -> do
                                         s <- mk a
-                                        return $ Skip (Right (s,t'))
+                                        s `seq` return (Skip (Right (s,t')))
                         Skip    t' -> return $ Skip (Left t')
                         Done       -> return $ Done
 
@@ -993,6 +1010,22 @@ unfoldrNM n f s = Stream step (s,n) (Max (delay_inline max n 0))
                                  Nothing     -> Done
                              ) (f s)
 
+-- | Apply monadic function n times to value. Zeroth element is original value.
+iterateNM :: Monad m => Int -> (a -> m a) -> a -> Stream m a
+{-# INLINE_STREAM iterateNM #-}
+iterateNM n f x0 = Stream step (x0,n) (Exact (delay_inline max n 0))
+  where
+    {-# INLINE_INNER step #-}
+    step (x,i) | i <= 0    = return Done
+               | i == n    = return $ Yield x (x,i-1)
+               | otherwise = do a <- f x
+                                return $ Yield a (a,i-1)
+
+-- | Apply function n times to value. Zeroth element is original value.
+iterateN :: Monad m => Int -> (a -> a) -> a -> Stream m a
+{-# INLINE_STREAM iterateN #-}
+iterateN n f x0 = iterateNM n (return . f) x0
+
 -- Scans
 -- -----
 
